package CDMA;

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.util.*;
import java.lang.Math;
import Library.*;
import Library.Geometry.*;

public class CDMA3D {
	CDMALayer                 parent;
	Library.Geometry.Observer observer;
	Library.Geometry.Observer planeObserver;
	V3D                       normal;
	C2D                       origin;
	int                       activeUser;
	CDMABar[][][]             signature;
	CDMABar[][]               amplitude;
	CDMABar[][]               transmitted;
	CDMABar[][]               received;
	double[][]                noise;
	int[]                     index;
	C2D                       separationLine;
	double                    combinedSignalsScale;

public CDMA3D(CDMALayer  parent) {
	this.parent = parent;
	R2D rect = parent.pb.getRectangle();
	observer = new Library.Geometry.Observer();
	planeObserver = new Library.Geometry.Observer();
	double scale = 0.2;
	observer.multiply(new M3D(scale,0,0,0,scale,0,0,0,scale));
	observer.setRotationSpeed(1.0);
	planeObserver.setRotationSpeed(1.0);
	planeObserver.setOrigin(new C3D(0,0,0));
	origin = new C2D((rect.w + parent.geomLeft)/2,rect.y + rect.h/2);
	activeUser = 0;

	reset();
	separationLine = new C2D(1,0);
	}

void reset() {
	createSignatures();
	createAmplitudes();
	createTransmitted();
	createReceived();
	createNoise();
	recompute();
	}

void recompute() {
	reconnectSignatures();
	reconnectAmplitudes();
	reconnectTransmitted();
	reconnectReceived();
	recomputeTransmitted();
	recomputeReceived();
	}

double randomBit() {
	if(Math.random()>=0.5) return 1;
	return -1;
	}

void createSignatures() {
	int N = parent.N;
	int L = parent.L;
	int M = parent.M;
	signature = new CDMABar[N][][];
	for(int n=0; n<N; n++) {
		signature[n] = new CDMABar[L][];
		for(int l=0; l<L; l++) {
			signature[n][l] = new CDMABar[M];
			for(int m=0; m<M; m++) {
				signature[n][l][m] = new CDMABar(this);
				signature[n][l][m].value = kronecker(n,l);
				signature[n][l][m].type  = CDMABar.signature;
				signature[n][l][m].user  = n;
				signature[n][l][m].item  = l;
				signature[n][l][m].index = m;
				};
			};
		};
	for(int i=0; i<L*L; i++) modifySignature();
	}

void reconnectSignatures() {
	int N = parent.N;
	int L = parent.L;
	int Q = parent.Q;
	for(int n=0; n<N; n++) {
		for(int l=0; l<L; l++) for(int q=0; q<Q; q++) {
			if(l>0)      connect(signature[n][l][q],signature[n][l-1][q]);
			else if(q>0) connect(signature[n][l][q],signature[n][L-1][q-1]);
			};
		connect(signature[n][0][0],null);
		connect(null,signature[n][L-1][Q-1]);
		};
	}

void createAmplitudes() {
	int N = parent.N;
	int M = parent.M;
	amplitude = new CDMABar[N][];
	for(int n=0; n<N; n++) {
		amplitude[n] = new CDMABar[M];
		for(int m=0; m<M; m++) {
			amplitude[n][m] = new CDMABar(this);
			amplitude[n][m].value = randomBit();
			amplitude[n][m].type  = CDMABar.amplitude;
			amplitude[n][m].user  = n;
			amplitude[n][m].index = m;
			};
		};
	}

void reconnectAmplitudes() {
	int N = parent.N;
	int Q = parent.Q;
	for(int n=0; n<N; n++) {
		for(int q=0; q<Q; q++) if(q>0) connect(amplitude[n][q],amplitude[n][q-1]);
		connect(amplitude[n][0],null);
		connect(null,amplitude[n][Q-1]);
		};
	}

void createTransmitted() {
	int L = parent.L;
	int M = parent.M;
	transmitted = new CDMABar[L][];
	for(int l=0; l<L; l++) {
		transmitted[l] = new CDMABar[M];
		for(int m=0; m<M; m++) {
			transmitted[l][m] = new CDMABar(this);
			transmitted[l][m].type  = CDMABar.transmitted;
			transmitted[l][m].item  = l;
			transmitted[l][m].index = m;
			};
		};
	}

void reconnectTransmitted() {
	int L = parent.L;
	int Q = parent.Q;
	for(int l=0; l<L; l++) for(int q=0; q<Q; q++) {
		if(l>0)      connect(transmitted[l][q],transmitted[l-1][q]);
		else if(q>0) connect(transmitted[l][q],transmitted[L-1][q-1]);
		};
	connect(transmitted[0][0],null);
	connect(null,transmitted[L-1][Q-1]);
	}

void createReceived() {
	int L = parent.L;
	int M = parent.M;
	received = new CDMABar[L][];
	for(int l=0; l<L; l++) {
		received[l] = new CDMABar[M];
		for(int m=0; m<M; m++) {
			received[l][m] = new CDMABar(this);
			received[l][m].type  = CDMABar.transmitted;
			received[l][m].item  = l;
			received[l][m].index = m;
			};
		};
	}

void reconnectReceived() {
	int L = parent.L;
	int Q = parent.Q;
	for(int l=0; l<L; l++) for(int q=0; q<Q; q++) {
		if(l>0)      connect(received[l][q],received[l-1][q]);
		else if(q>0) connect(received[l][q],received[L-1][q-1]);
		};
	connect(received[0][0],null);
	connect(null,received[L-1][Q-1]);
	}

void createNoise() {
	int L = parent.L;
	int M = parent.M;
	noise  = new double[L][];
	Random rand = new Random();
	for(int l=0; l<L; l++) {
		noise[l]  = new double[M];
		for(int m=0; m<M; m++) noise[l][m] = rand.nextGaussian();
		};
	}

void recomputeTransmitted() {
	int N = parent.N;
	int L = parent.L;
	int M = parent.M;
	for(int l=0; l<L; l++) for(int m=0; m<M; m++) {
		transmitted[l][m].value = 0;
		for(int n=0; n<N; n++)
			transmitted[l][m].value += signature[n][l][m].value * amplitude[n][m].value;
		};
	}

void recomputeReceived() {
	int L = parent.L;
	int M = parent.M;
	double noiseMagnitude = parent.noiseMagnitude();
	for(int l=0; l<L; l++) for(int m=0; m<M; m++) {
		received[l][m].value = transmitted[l][m].value;
		received[l][m].value += noiseMagnitude * noise[l][m];
		};
	}

void modifySignature(int l1, int l2, double alpha) {
	int N = parent.N;
	int L = parent.L;
	int M = parent.M;
	double sin = Math.sin(alpha);
	double cos = Math.cos(alpha);
	for(int n=0; n<N; n++) for(int m=0; m<M; m++) {
		double s1 = signature[n][l1][m].value;
		double s2 = signature[n][l2][m].value;
		signature[n][l1][m].value = cos*s1 - sin*s2;
		signature[n][l2][m].value = sin*s1 + cos*s2;
		};
	}

void modifySignature() {
	int L = parent.L;
	int l1 = (int) (L*Math.random());
	int l2 = l1;
	while(l2 == l1) l2 = (int) (L*Math.random());
	double alpha = 2*Math.PI*Math.random();
	modifySignature(l1,l2,alpha);
	}

void connect(CDMABar b1, CDMABar b2) {
	if(b2!=null) b2.next = b1; if(b1!=null) b1.pred = b2;
	}

double kronecker(int i, int j) {if(i==j) return 1.0; return 0.0;}

C2D transformToScreen(V3D c) {
	V3D ct = observer.transformVector(c);
	return projectToScreen(ct);
	}

C2D projectToScreen(V3D c) {
	return new C2D(c.x+origin.x,c.y+origin.y);
	}

void paintAxis2D(Paintbrush pb,int k) {
	C2D from=null, upto=null;
	char name = ' ';
	double d = 0.3;
	if(k==0) {
		from = new C2D(-d+origin.x,0+origin.y);
		upto = new C2D(d+origin.x,0+origin.y);
		name = 'x';
		};
	if(k==1) {
		from = new C2D(0+origin.x,-d+origin.y);
		upto = new C2D(0+origin.x,d+origin.y);
		name = 'y';
		};
	pb.setColor(Color.black);
	pb.drawLine(from,upto);
	pb.setFont(0.05);
	pb.drawString(""+name,upto);
	}

void paintAxis3D(Paintbrush pb,int k) {
	V3D from3 = null, upto3 = null;
	C2D from2, upto2;
	char name = ' ';
	if(k==0) {from3 = new V3D(1,0,0); upto3 = new V3D(-1,0,0); name = 'x';};
	if(k==1) {from3 = new V3D(0,1,0); upto3 = new V3D(0,-1,0); name = 'y';};
	if(k==2) {from3 = new V3D(0,0,1); upto3 = new V3D(0,0,-1); name = 'z';};
	from2 = transformToScreen(from3);
	upto2 = transformToScreen(upto3);
	pb.setColor(Color.black);
	pb.drawLine(from2,upto2);
	pb.setFont(0.05);
	pb.drawString(""+name,from2);
	}

V3D add(V3D a, V3D b) {return new V3D(a.x+b.x,a.y+b.y,a.z+b.z);}
V3D mul(V3D a, double m) {return new V3D(m*a.x,m*a.y,m*a.z);}

void paintPlane(Paintbrush pb) {
	V3D np = new V3D(normal.y,-normal.x,0);
	V3D vp = V3D.vectorProduct(normal,np);
	V3D[] a  = new V3D[4];
	double w = 2;
	a[0] = new V3D( w, w,0);
	a[1] = new V3D( w,-w,0);
	a[3] = new V3D(-w, w,0);
	a[2] = new V3D(-w,-w,0);
	if(vp.length() > 0.000001) {
		np.normalize();
		vp.normalize();
		np = mul(np,w);
		vp = mul(vp,w);
		a[0] = add(np,        vp);
		a[1] = add(np,        mul(vp,-1));
		a[3] = add(mul(np,-1),vp);
		a[2] = add(mul(np,-1),mul(vp,-1));
		};
	C2D[] point = new C2D[4];
	for(int i=0; i<4; i++) point[i] = transformToScreen(a[i]);
	StandardPolygon plane = new StandardPolygon(point);
	int acolor = 180, bcolor = 120;
	Color lightColor = new Color(acolor,acolor,255);
	Color darkColor  = new Color(bcolor,bcolor,255);
//p.p("----------------");
	Color color = null;
	if(observer.transformVector(normal).z > 0) color = lightColor; else color = darkColor;
	paintVectors3D(pb,false);
	pb.setColor(color);
	pb.fillPolygon(plane);
	pb.setColor(Color.black);
	pb.drawPolygon(plane);	
	paintVectors3D(pb,true);
	}

Color sampleColor(int j) {
	if(j == parent.currentSample) return Color.yellow;
	if(amplitude[activeUser][j].value > 0) return Color.black;
	return Color.red;
	}

void paintVector2D(Paintbrush pb, C2D v, Color c) {
	pb.setColor(c);
	StandardArrow arrow = new StandardArrow(origin,v,StandardArrow.ARROW,0.01,0.05,0.02);
	pb.fillPolygon(arrow);
	}

void paintVectors2D(Paintbrush pb) {
	int N = parent.N;
	Color c = null;
	double scale = 0.4;
	for(int n=0; n<2; n++) {
		C2D s = new C2D(signature[n][0][0].value,signature[n][1][0].value);
		C2D head = new C2D(transform2D(s));
		if(n==0) c = new Color(255,0,0);
		if(n==1) c = new Color(0,255,0);
		paintVector2D(pb,head,c);
		};
	if(parent.guide.scene == parent.guide.samples2D) if(parent.currentSample > 0) {
		C2D s = transform2D(receivedC2D(parent.currentSample));
		paintVector2D(pb,s,Color.black);
		};
	}

void paintNormal2D(Paintbrush pb) {
	if(parent.guide.scene == parent.guide.decoding2D) {
		double dx = separationLine.x-origin.x;
		double dy = separationLine.y-origin.y;
		double l = Math.sqrt(dx*dx+dy*dy);
		if(l != 0) {dx /= 3*l; dy /= 3*l;};
		C2D head = new C2D(dy+origin.x,-dx+origin.y);
		paintVector2D(pb,head,new Color(255,255,127));
		};
	if(parent.currentSample > 0) {
		C2D c = transform2D(receivedC2D(parent.currentSample));
		paintVector2D(pb,c,Color.black);
		};
	}

boolean isInFrontOfPlane(V3D v, boolean inFrontOfPlane) {
	V3D n = observer.transformVector(normal);
	if(n.z >= 0) {
		if(inFrontOfPlane) return V3D.scalarProduct(v,n) > 0;
		else return V3D.scalarProduct(v,n) <= 0;
		}
	else	{
		if(inFrontOfPlane) return V3D.scalarProduct(v,n) < 0;
		else return V3D.scalarProduct(v,n) >= 0;
		}
	}

void paintVector3D(Paintbrush pb, V3D v, Color c, boolean inFrontOfPlane) {
	V3D head3 = observer.transformVector(v);
	 if(isInFrontOfPlane(head3,inFrontOfPlane)) {
		C2D head2 = projectToScreen(head3);
		paintVector2D(pb,head2,c);
		};
	}

void paintVectors3D(Paintbrush pb, boolean inFrontOfPlane) {
	Color c = null;
	if(parent.control.showVectors.getState()) for(int n=0; n<3; n++) {
		V3D head = new V3D(signature[n][0][0].value,signature[n][1][0].value,signature[n][2][0].value);
		if(n==0) c = new Color(255,0,0);
		if(n==1) c = new Color(0,255,0);
		if(n==2) c = new Color(0,0,255);
		paintVector3D(pb,head,c,inFrontOfPlane);
		};
	if(parent.control.showNormal.getState()) {
		Color lightNormalColor = new Color(255,255,0);
		Color darkNormalColor  = new Color(196,196,0);
		paintVector3D(pb,normal, lightNormalColor,inFrontOfPlane);
		V3D nnormal = new V3D(-normal.x,-normal.y,-normal.z);
		paintVector3D(pb,nnormal,darkNormalColor, inFrontOfPlane);
		if(parent.currentSample>0) {
			paintVector3D(pb,receivedV3D(parent.currentSample),Color.black,inFrontOfPlane);
			};
		};
	if(parent.control.showVector.getState()) {
		paintVector3D(pb,transmittedV3D(0),Color.black,inFrontOfPlane);
		};
	}

void paintCombinedVector2D(Paintbrush pb) {
//	double scale = 0.4;
//	double x = scale * transmitted[0][0].value+origin.x;
//	double y = scale * transmitted[1][0].value+origin.y;
	C2D head = transform2D(transmittedC2D(0));
	pb.setColor(new Color(0,0,0));
	StandardArrow arrow = new StandardArrow(origin,head,StandardArrow.ARROW,0.01,0.05,0.02);
	pb.fillPolygon(arrow);
	}

void paintSignaturesBackground(Paintbrush pb) {
	int g = 240;
	pb.setColor(new Color(g,g,255));
	pb.fillRect(parent.geomLeft,0,2,1);
	}

boolean showAmplitudes() {
	if(parent.control.showAmplitudes.getState()) return true;
	if(parent.currentSample >= 0) return true;
	return false;
	}

void separate(C2D mousePosition) {
	if((mousePosition.x != 0) || (mousePosition.y != 0)) separationLine = mousePosition;
	}

C2D transform2D(C2D c) {
	double scale = 0.25;
	return new C2D(scale*c.x+origin.x,scale*c.y+origin.y);
	}

C2D transmittedC2D(int m) {return new C2D(transmitted[0][m].value,transmitted[1][m].value);}
V3D transmittedV3D(int m) {
	double x = transmitted[0][m].value;
	double y = transmitted[1][m].value;
	double z = transmitted[2][m].value;
	return new V3D(x,y,z);
	}

C2D receivedC2D(int m) {return new C2D(received[0][m].value,received[1][m].value);}
V3D receivedV3D(int m) {
	double x = received[0][m].value;
	double y = received[1][m].value;
	double z = received[2][m].value;
	return new V3D(x,y,z);
	}


public void paint2D(Paintbrush pb) {
	int N = parent.N;
	int L = parent.L;
	int M = parent.M;
	R2D pbr  = pb.getRectangle();
	normal   = planeObserver.transformVector(new V3D(0,0,1));
	V3D norm = observer.transformVector(normal);

	if(parent.control.showSignatures.getState()) paintSignaturesBackground(pb);
	if(parent.control.showPoints.getState()) for(int m=1; m<M; m++) {
		C2D s = transform2D(receivedC2D(m));
		pb.setColor(sampleColor(m));
		pb.fillCircle(s,parent.P.sampleR);
		};
//	paintPlane(pb);
	for(int k=0; k<2; k++) if(parent.control.showAxes.getState()) paintAxis2D(pb,k);
	if(parent.guide.scene == parent.guide.decoding2D) {
		pb.setColor(Color.blue);
		double dx = separationLine.x - origin.x;
		double dy = separationLine.y - origin.y;
		double m = 25;
		C2D p1 = new C2D(origin.x+m*dx,origin.y+m*dy);
		C2D p2 = new C2D(origin.x-m*dx,origin.y-m*dy);
		pb.drawLine(p1,p2);
		};
	pb.setColor(new Color(196,196,255));
	pb.fillRect(pbr.x,pbr.y,parent.geomLeft-pbr.x,pbr.h);
	if(parent.control.showSignatures.getState())
		for(int n=0; n<N; n++) for(int l=0; l<L; l++) {
			signature[n][l][0].paintRect(pb);
			};
	if(showAmplitudes()) {
		for(int n=0; n<N; n++) {
			amplitude[n][0].paintRect(pb);
			};
		for(int l=0; l<L; l++) {
			transmitted[l][0].paintRect(pb);
			};
		};
	if(parent.control.showSignatures.getState())
		for(int n=0; n<N; n++) for(int l=0; l<L; l++) {
			signature[n][l][0].paint(pb);
			};
	if(showAmplitudes()) {
		for(int n=0; n<N; n++) {
			amplitude[n][0].paint(pb);
			};
		for(int l=0; l<L; l++) {
			transmitted[l][0].paint(pb);
			};
		};
	if(parent.control.showVectors.getState()) paintVectors2D(pb);
	if(parent.control.showNormal.getState())  paintNormal2D(pb);
	if(parent.control.showVector.getState ()) paintCombinedVector2D(pb);

	paintLegend(pb);
	}

public void paint3D(Paintbrush pb) {
	int N = parent.N;
	int L = parent.L;
	int M = parent.M;
	R2D pbr  = pb.getRectangle();
	normal   = planeObserver.transformVector(new V3D(0,0,1));
	V3D norm = observer.transformVector(normal);

	if(parent.control.showSignatures.getState()) paintSignaturesBackground(pb);
	V3D[] transformedSample = new V3D[M];
	index = new int[M];
	for(int m=1; m<M; m++) {
		transformedSample[m] = observer.transformVector(receivedV3D(m));
		index[m] = m;
		};
	for(int i=2; i<transformedSample.length; i++) for(int j=1; j<i; j++) {
		if(transformedSample[index[i]].z < transformedSample[index[j]].z) {
			int l = index[i]; index[i] = index[j]; index[j] = l;
			};
		};
	for(int i=1; i<M; i++) if(parent.control.showPoints.getState()) {
		int j = index[i];
		C2D s = projectToScreen(transformedSample[j]);
		pb.setColor(sampleColor(j));
		if(V3D.scalarProduct(norm,transformedSample[j])*norm.z < 0)
			pb.fillCircle(s,parent.P.sampleR);
		};
	if(parent.control.showPlane.getState()) paintPlane(pb); else paintVectors3D(pb,true);
	for(int m=1; m<M; m++) if(parent.control.showPoints.getState()) {
		int j = index[m];
		C2D s = projectToScreen(transformedSample[j]);
		pb.setColor(sampleColor(j));
		if(V3D.scalarProduct(norm,transformedSample[j])*norm.z >= 0)
			pb.fillCircle(s,parent.P.sampleR);
		};
	for(int k=0; k<3; k++) if(parent.control.showAxes.getState()) paintAxis3D(pb,k);
	pb.setColor(new Color(196,196,255));
	pb.fillRect(pbr.x,pbr.y,parent.geomLeft-pbr.x,pbr.h);
	if(parent.control.showSignatures.getState())
		for(int n=0; n<N; n++) for(int l=0; l<L; l++) {
			signature[n][l][0].paintRect(pb);
			};
	if(showAmplitudes()) {
		for(int n=0; n<N; n++) {
			amplitude[n][0].paintRect(pb);
			};
		for(int l=0; l<L; l++) {
			transmitted[l][0].paintRect(pb);
			};
		};
	if(parent.control.showSignatures.getState())
		for(int n=0; n<N; n++) for(int l=0; l<L; l++) {
			signature[n][l][0].paint(pb);
			};
	if(showAmplitudes()) {
		for(int n=0; n<N; n++) {
			amplitude[n][0].paint(pb);
			};
		for(int l=0; l<L; l++) {
			transmitted[l][0].paint(pb);
			};
		};

	paintLegend(pb);
	}

public void paintPlot(Paintbrush pb) {
	int N = parent.N;
	int L = parent.L;
	int Q = parent.Q;
	for(int n=0; n<N; n++) for(int l=0; l<L; l++) for(int q=0; q<Q; q++) 
		signature[n][l][q].paintRect(pb);
	for(int n=0; n<N; n++) for(int l=0; l<L; l++) for(int q=0; q<Q; q++) 
		signature[n][l][q].paint(pb);
	if(parent.control.showAmplitudes.getState()) {
		for(int n=0; n<N; n++) for(int q=0; q<Q; q++)
			amplitude[n][q].paintRect(pb);
		for(int l=0; l<L; l++) for(int q=0; q<Q; q++)
			transmitted[l][q].paintRect(pb);
		for(int n=0; n<N; n++) for(int q=0; q<Q; q++)
			amplitude[n][q].paint(pb);
		for(int l=0; l<L; l++) for(int q=0; q<Q; q++)
			transmitted[l][q].paint(pb);
		};
	}

void paintLegend(Paintbrush pb) {
	int N = parent.N;
	pb.setColor(Color.black);
	double x = 0.025;
	double fontSize = 0.075 / N;
	double fontSizeMax = 0.03;
	if(fontSize > fontSizeMax) fontSize = fontSizeMax;
	pb.setFont(fontSize);
	for(int n=0; n<N; n++) {
		pb.drawString("User "+n,   x, parent.plot.sgny + 2*n*parent.plot.spacey);
		pb.drawString("Signature", x, parent.plot.sgny + 2*n*parent.plot.spacey + fontSize);
		};
	if(showAmplitudes()) {
		for(int n=0; n<N; n++) {
			pb.drawString("User "+n,   x, parent.plot.sgny + (2*n+1)*parent.plot.spacey);
			pb.drawString("Signal",    x, parent.plot.sgny + (2*n+1)*parent.plot.spacey + fontSize);
			};
		pb.drawString("Combined", x, parent.plot.sigy);
		pb.drawString("signal",   x, parent.plot.sigy + fontSize);
		};
	}

public void paint(Paintbrush pb) {
	paintLegend(pb);
	recompute();
	if     (parent.guide.plot())  paintPlot(pb);
	else if(parent.guide.geo2D()) paint2D(pb);
	else                          paint3D(pb);
	}
}

