package Delis;

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import java.lang.Math;
import java.util.Random;
import Library.*;
import Library.History.*;
import Library.Geometry.*;

public class WirelessNetwork {
	DelisLayer layer;
	WirelessNode[] node;
	WirelessEdge firstEdge;
	boolean edgesInvalid;
	boolean showTransmit;
	boolean showEdges;
	boolean drawWideEdges;
	WirelessEdge firstBridge;
	WirelessLattice lattice;

	double[][] distance;
	int[][]    direction;

	R2D paintRectangle;

	int phase;


public WirelessNetwork(DelisLayer layer) {
	this.layer = layer;
	firstEdge = null;
	firstBridge = null;
	edgesInvalid = true;
	showTransmit = layer.control.showTransmit.getState();
	showEdges = layer.control.showEdges.getState();
	drawWideEdges = showTransmit;
	newLattice();
	}

void newLattice() {
	int latticeNboxes = layer.control.latticeBoxField. getIntValue();
	lattice = new WirelessLattice(this,latticeNboxes);
	}

public int size() {if(node == null) return 0; return node.length;}

public int largestComponentSize() {
	int result = 0;
	for(int i=0; i<node.length; i++) {
		WirelessNode n = node[i];
		if(n.componentSuperior == n) if(result < n.componentSize) result = n.componentSize;
		};
	return result;
	}

double determinePower() {
	double result = 0;
	for(int i=0; i<node.length; i++) result += node[i].transmitRadius*node[i].transmitRadius;
	return result;
	}

double determineCongestion() {
	int[][] congestion;
	congestion = new int[node.length][];
	for(int i=0; i<node.length; i++) {
		congestion[i] = new int[node.length];
		for(int j=0; j<node.length; j++) congestion[i][j] = 0;
		};
	for(int i=0; i<node.length; i++) for(int j=0; j<node.length; j++) {
		int p1 = i;
		while(p1 != j) {
			int p2 = direction[p1][j];
			congestion[p1][p2] ++;
			p1 = p2;
			};
		};
	int sum = 0;
	for(int i=0; i<node.length; i++) for(int j=0; j<node.length; j++) sum += congestion[i][j];
	double result = sum;
	result /= node.length*node.length;
	return result;
	}

void computeDistance() {			// Floyd-Warshall algorithm
	int n = size();				// initialize basic variables
	distance  = new double[n][];
	direction = new int[n][];
	for(int i=0; i<n; i++) {		// initialize the distance and direction fields
		distance[i]  = new double[n];
		direction[i] = new int[n];
		for(int j=0; j<n; j++) {
			distance[i][j]  = 1E12;
			direction[i][j] = j;
			};
		};
	WirelessEdge e = firstEdge;		// set edge distances
	while(e != null) {
		int i1 = e.n1.ID;
		int i2 = e.n2.ID;
		distance[i1][i2] = e.length();
		distance[i2][i1] = distance[i1][i2];
		e = e.next;
		};
	e = firstBridge;			// set bridge distances
	while(e != null) {
		int i1 = e.n1.ID;
		int i2 = e.n2.ID;
		distance[i1][i2] = e.length();
		distance[i2][i1] = distance[i1][i2];
		e = e.next;
		};
						// the computation according Floyd-Warshall
	 for(int k=0; k<n; k++) for(int i=0; i<n; i++) for(int j=0; j<n; j++) {
		double d = distance[i][k] + distance[k][j];
		if(distance[i][j] > d) {
			distance[i][j]  = d;
			direction[i][j] = direction[i][k];
			};
		};
	}

WirelessEdge findBridge() {
	WirelessEdge worstPair = null;
	double worstRatio = 0;
	for(int i=1; i<size(); i++) for(int j=0; j<i; j++) {
		double ratio = distance[i][j];
		ratio /= C2D.distance(node[i].position,node[j].position);
		if(ratio > 1.02) if((worstPair == null) || (worstRatio < ratio)) {
			worstPair = new WirelessEdge(this,node[i],node[j]);
			worstPair.thick = true;
			worstPair.automaticColor = false;
			worstRatio = ratio;
			};
		};
	return worstPair;
	}

void setTransmitRadii() {
	Screen screen = layer.screen;
	History history = screen.history;
	double[] newRadius = new double[node.length];
	for(int i=0; i<node.length; i++) newRadius[i] = 0;
	for(int i=0; i<node.length; i++) for(int j=0; j<node[i].degree; j++) {
		WirelessNode ngbh = node[i].neighbor[j];
		double d = C2D.distance(node[i].position,ngbh.position);
		if(newRadius[i] < d) newRadius[i] = d;
		if(newRadius[ngbh.ID] < d) newRadius[ngbh.ID] = d;
		};
	for(int i=0; i<node.length; i++) if(node[i].transmitRadius !=newRadius[i]) 
		history.appendex(new WirelessNodeRadiusChange(history,layer,node[i],newRadius[i]));
	determineEdges();
	}

void changeRadius(boolean increase) {
	if(increase) if(connected()) return;
	if(!increase) if(largestRadius() <= layer.P.nodeLargeRadius) return;
	History history = layer.screen.history;
	history.clear();
	history.breakPoint(History.END_LEVEL);
	if(layer.proportionalRadius()) {
		double scale = 1; scale /= (1+30*largestRadius()); scale += 1;
		if(!increase) scale = 1.0/scale;
		for(int i=0; i<node.length; i++) node[i].changeRadius(scale);
		}
	else if(layer.nngRadius()) {
		initializeComponents();
		for(int i=0; i<node.length; i++) for(int j=0; j<node[i].degree; j++) 
			join(node[i],node[i].neighbor[j]);
		for(int i=0; i<node.length; i++) node[i].changeNeighbors(increase);
		setTransmitRadii();
		};
	history.breakPoint(History.END_LEVEL);
	history.rewind();
	}

double largestRadius() {
	double r = 0;
	for(int i=0; i<size(); i++) if(r < node[i].transmitRadius) r = node[i].transmitRadius;
	return r;
	}

boolean connected() {return node[0].componentSize() == size();}

void initializeComponents() {
	for(int i= 0; i<node.length; i++) node[i].initializeComponent();
	}

void join(WirelessNode n1, WirelessNode n2) {
	WirelessNode m1 = n1.componentRepresentative();
	WirelessNode m2 = n2.componentRepresentative();
	if(m1 != m2) {
		if(m1.componentSize >= m2.componentSize) {
			m1.componentSize += m2.componentSize;
			m2.componentSuperior = m1;
			}
		else	{
			m2.componentSize += m1.componentSize;
			m1.componentSuperior = m2;
			};
		};
	}

void appendEdge(WirelessEdge e) {
	e.next = firstEdge;
	firstEdge = e;
	join(e.n1,e.n2);
	}

void appendBridge(WirelessEdge e) {
	if(layer.nBridges == 0) {
		layer.powerCurve.setValue(0,determinePower());
		layer.congestionCurve.setValue(0,determineCongestion());
		};
	e.next = firstBridge;
	firstBridge = e;
	join(e.n1,e.n2);
	double d = C2D.distance(e.n1.position,e.n2.position);
	if(e.n1.transmitRadius < d) e.n1.transmitRadius = d;
	if(e.n2.transmitRadius < d) e.n2.transmitRadius = d;
	determineEdges();
	layer.nBridges ++;
	if((layer.nBridges * layer.powerPlot.xScale) >= 1) {
		layer.powerPlot.xScale /= 2;
		layer.powerCurveUniform.setValue(1.0/layer.powerPlot.xScale,layer.uniformPower);
		layer.congestionPlot.xScale /= 2;
		layer.congestionCurveUniform.setValue(1.0/layer.congestionPlot.xScale,layer.uniformCongestion);
		};
	layer.powerCurve.     setValue(layer.nBridges,determinePower());
	layer.congestionCurve.setValue(layer.nBridges,determineCongestion());
phase = 0;
	}

void determineEdges() {
	firstEdge = null;
	
	for(int i=1; i<node.length; i++) for(int j=0; j<i; j++) {
		double d = 0.999*C2D.distance(node[i].position,node[j].position);
		if((node[i].transmitRadius >= d) && (node[j].transmitRadius >= d))
			appendEdge(new WirelessEdge(this,node[i],node[j]));
		};
	edgesInvalid = false;
	}

public double dimension2paint(double x) {return x * paintRectangle.w;}

public C2D position2paint(C2D position) {
	double x = paintRectangle.x + position.x * paintRectangle.w;
	double y = paintRectangle.y + position.y * paintRectangle.h;
	return new C2D(x,y);
	}

public R2D rectangle2paint(R2D rectangle) {
	double x = paintRectangle.x + rectangle.x * paintRectangle.w;
	double y = paintRectangle.y + rectangle.y * paintRectangle.h;
	double w = rectangle.w * paintRectangle.w;
	double h = rectangle.h * paintRectangle.h;
	return new R2D(x,y,w,h);
	}

public void paintBackground(Paintbrush pb) {
	int g = 240;
	pb.setColor(g,g,g);
	pb.fillRect(paintRectangle);
	}

public void paint(Paintbrush pb) {
	if(edgesInvalid) {initializeComponents(); determineEdges();};
	if(layer.control.showLattice.checked()) lattice.paint(pb);
	if(firstBridge != null) if(layer.guide.scene == layer.guide.bridgeScream) {
		WirelessNode center = firstBridge.n2;
		double r = dimension2paint(C2D.distance(center.position,firstBridge.n1.position));
		C2D pos  = position2paint(center.position);
		pb.setColor(new Color(255,255,128));
		pb.fillCircle(pos,r);
		pb.setColor(new Color(128,128,128));
		pb.drawCircle(pos,r);
		};
	if(node != null) if(showTransmit) {
		for(int i=0; i<node.length; i++) node[i].paintTransmit(pb);
		for(int i=0; i<node.length; i++) node[i].paintBoundary(pb);
		};
	WirelessEdge e = firstEdge;
	if(showEdges) while(e != null) {e.paint(pb); e = e.next;};
	WirelessEdge b;
	b = firstBridge;
	while(b != null) {
		if(b == firstBridge) b.color = Color.red;
		else                 b.color = new Color(255,0,255);
		b.paint(pb);
		b = b.next;
		};
	b = firstBridge;
	while(b != null) {
		int p1 = b.n1.ID;
		int ph = 0;
		if(b == firstBridge) while(p1 != b.n2.ID) {
			int p2 = direction[p1][b.n2.ID];
			pb.setColor(Color.green);
			if((ph++ < phase) || (layer.guide.scene != layer.guide.bridgeScream)) {
				C2D pos1 = position2paint(node[p1].position);
				C2D pos2 = position2paint(node[p2].position);
				double r = dimension2paint(node[p1].nodeLargeRadius)/2;
				pb.drawThickLine(pos1,pos2,r);
				};
			p1 = p2;
			};
		b = b.next;
		};
	if(node != null) for(int i=0; i<node.length; i++) node[i].paintNode(pb);
	}


}

